"""Base functionality for async communication.

Distributed under the GNU General Public License v2
Copyright (C) 2024 Alex Ruddick
Copyright (C) 2023 NuMat Technologies
"""
from __future__ import annotations

import asyncio
import logging
from abc import ABC, abstractmethod
from typing import Any

import serial

logger = logging.getLogger('alicat')


class Client(ABC):
    """Serial or TCP client."""

    def __init__(self, timeout: float):
        """Initialize common attributes."""
        self.address = ''
        self.open = False
        self.timeout = timeout
        self.timeouts = 0
        self.max_timeouts = 10
        self.connection: dict[str, Any] = {}
        self.reconnecting = False
        self.eol = b'\r'
        self.lock = asyncio.Lock()

    @abstractmethod
    async def _write(self, message: str) -> None:
        """Write a message to the device."""
        ...

    @abstractmethod
    async def _read(self, length: int) -> str:
        """Read a fixed number of bytes from the device."""

    @abstractmethod
    async def _readline(self) -> str:
        """Read until a LF terminator."""

    async def _write_and_read(self, command: str) -> str | None:
        """Write a command and read a response.

        As industrial devices are commonly unplugged, this has been expanded to
        handle recovering from disconnects.
        """
        await self._handle_connection()
        async with self.lock:
            if self.open:
                try:
                    response = await self._handle_communication(command)
                    return response
                except asyncio.exceptions.IncompleteReadError:
                    logger.error('IncompleteReadError.  Are there multiple connections?')
                    return None
            else:
                return None

    async def _clear(self) -> None:
        """Clear the reader stream when it has been corrupted from multiple connections."""
        logger.warning("Multiple connections detected; clearing reader stream.")
        try:
            junk = await asyncio.wait_for(self._read(100), timeout=0.5)
            logger.warning(junk)
        except TimeoutError:
            pass

    async def _handle_communication(self, command: str) -> str | None:
        """Manage communication, including timeouts and logging."""
        try:
            await self._write(command)
            result = await self._readline()
            self.timeouts = 0
            return result
        except (asyncio.TimeoutError, TypeError, OSError):
            self.timeouts += 1
            if self.timeouts == self.max_timeouts:
                logger.error(f'Reading from {self.address} timed out '
                             f'{self.timeouts} times.')
                await self.close()
            return None

    @abstractmethod
    async def _handle_connection(self) -> None:
        pass

    @abstractmethod
    async def close(self) -> None:
        """Close the connection."""
        pass


class TcpClient(Client):
    """A generic reconnecting asyncio TCP client.

    This base functionality can be used by any industrial control device
    communicating over TCP.
    """

    def __init__(self, address: str, timeout: float=1.0):
        """Communicator using a TCP/IP<=>serial gateway."""
        super().__init__(timeout)
        try:
            self.address, self.port = address.split(':')
        except ValueError as e:
            raise ValueError('address must be hostname:port') from e

    async def __aenter__(self) -> Client:
        """Provide async entrance to context manager.

        Contrasting synchronous access, this will connect on initialization.
        """
        await self._handle_connection()
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Provide async exit to context manager."""
        await self.close()

    async def _connect(self) -> None:
        """Asynchronously open a TCP connection with the server."""
        await self.close()
        reader, writer = await asyncio.open_connection(self.address, self.port)
        self.connection = {'reader': reader, 'writer': writer}
        self.open = True

    async def _read(self, length: int) -> str:
        """Read a fixed number of bytes from the device."""
        await self._handle_connection()
        response = await self.connection['reader'].read(length)
        return response.decode().strip()

    async def _readline(self) -> str:
        """Read until a line terminator."""
        await self._handle_connection()
        response = await self.connection['reader'].readuntil(self.eol)
        return response.decode().strip().replace('\x00', '')

    async def _write(self, command: str) -> None:
        """Write a command and do not expect a response.

        As industrial devices are commonly unplugged, this has been expanded to
        handle recovering from disconnects.
        """
        await self._handle_connection()
        self.connection['writer'].write(command.encode() + self.eol)

    async def _handle_connection(self) -> None:
        """Automatically maintain TCP connection."""
        if self.open:
            return
        try:
            await asyncio.wait_for(self._connect(), timeout=0.75)
            self.reconnecting = False
        except (asyncio.TimeoutError, OSError):
            if not self.reconnecting:
                logger.error(f'Connecting to {self.address} timed out.')
            self.reconnecting = True

    async def _handle_communication(self, command: str) -> str | None:
        """Manage communication, including timeouts and logging."""
        try:
            await self._write(command)
            future = self._readline()
            result = await asyncio.wait_for(future, timeout=0.75)
            self.timeouts = 0
            return result
        except (asyncio.TimeoutError, TypeError, OSError):
            self.timeouts += 1
            if self.timeouts == self.max_timeouts:
                logger.error(f'Reading from {self.address} timed out '
                             f'{self.timeouts} times.')
                await self.close()
            return None

    async def close(self) -> None:
        """Close the TCP connection."""
        if self.open:
            self.connection['writer'].close()
            await self.connection['writer'].wait_closed()
        self.open = False


class SerialClient(Client):
    """Client using a directly-connected RS232 serial device."""

    def __init__(self, address: str, baudrate: int=19200, timeout: float=.15,
                 bytesize: int = serial.EIGHTBITS,
                 stopbits: int = serial.STOPBITS_ONE,
                 parity: str = serial.PARITY_NONE):
        """Initialize serial port."""
        super().__init__(timeout)
        self.address = address
        assert isinstance(self.address, str)
        self.serial_details = {'baudrate': baudrate,
                               'bytesize': bytesize,
                               'stopbits': stopbits,
                               'parity': parity,
                               'timeout': timeout}
        self.ser = serial.Serial(self.address, **self.serial_details)  # type: ignore [arg-type]

    async def _read(self, length: int) -> str:
        """Read a fixed number of bytes from the device."""
        return self.ser.read(length).decode()

    async def _readline(self) -> str:
        """Read until a LF terminator."""
        return self.ser.readline().strip().decode().replace('\x00', '')

    async def _write(self, message: str) -> None:
        """Write a message to the device."""
        self.ser.write(message.encode() + self.eol)

    async def close(self) -> None:
        """Release resources."""
        self.ser.close()

    async def _handle_connection(self) -> None:
        self.open = True

def _is_float(msg: Any) -> bool:
    try:
        float(msg)
        return True
    except ValueError:
        return False



"""Test the driver responds with correct data."""
from random import uniform
from unittest import mock

import pytest

from alicat import command_line

# from alicat.driver import FlowController
from alicat.mock import FlowController

ADDRESS = '/dev/tty.usbserial-FTCJ5EK9'


@pytest.mark.parametrize('unit', ['A', 'B'])
@mock.patch('alicat.FlowController', FlowController)
def test_driver_cli(capsys, unit):
    """Confirm the commandline interface works with different unit IDs."""
    command_line([ADDRESS, '--unit', unit])
    captured = capsys.readouterr()
    assert ("mass_flow" in captured.out)


async def test_flow_setpoint_roundtrip():
    """Confirm that setting/getting flowrates works."""
    async with FlowController(ADDRESS) as device:
        flow_sp = round(uniform(0.01, 0.1), 2)
        await device.set_flow_rate(flowrate=flow_sp)
        # assert flow_sp == await device.get_flow_rate()
        result = await device.get()
        assert flow_sp == result['setpoint']


async def test_lock_unlock():
    """Confirm that locking/unlocking the buttons works."""
    async with FlowController(ADDRESS) as device:
        await device.lock()
        assert await device.is_locked()
        await device.unlock()
        assert not await device.is_locked()


@pytest.mark.parametrize('gas', ['Air', 'H2'])
async def test_set_standard_gas_name(gas):
    """Confirm that setting standard gases by name works."""
    async with FlowController(ADDRESS) as device:
        await device.set_gas(gas)
        result = await device.get()
        assert gas == result['gas']


@pytest.mark.parametrize('gas', [('Air', 0), ('H2', 6)])
async def test_set_standard_gas_number(gas):
    """Confirm that setting standard gases by number works."""
    async with FlowController(ADDRESS) as device:
        await device.set_gas(gas[1])
        result = await device.get()
        assert gas[0] == result['gas']


async def test_get_firmware():
    """Confirm the firmware version can be read."""
    async with FlowController(ADDRESS) as device:
        result = await device.get_firmware()
        assert 'v' in result or 'GP' in result


@pytest.mark.parametrize('config', [
    {'up': True, 'down': False, 'zero': True, 'power': False},
    {'up': True, 'down': True, 'zero': False, 'power': True},
    {'up': False, 'down': False, 'zero': False, 'power': False},
    ])
async def test_ramp_config(config):
    """Confirm changing the ramping configuration works."""
    async with FlowController(ADDRESS) as device:
        await device.set_ramp_config(config)
        result = await device.get_ramp_config()
        assert config == result



"""Python driver for Alicat mass flow controllers.

Distributed under the GNU General Public License v2
Copyright (C) 2023 NuMat Technologies
"""
from typing import Any

from alicat.driver import FlowController, FlowMeter  # noqa


def command_line(args: Any = None) -> None:
    """CLI interface, accessible when installed through pip."""
    import argparse
    import asyncio
    import json
    from time import time
    parser = argparse.ArgumentParser(description="Control an Alicat mass "
                                     "flow controller from the command line.")
    parser.add_argument('address', type=str, default='/dev/ttyUSB0', help="The "
                        "target serial port (or TCP address:port). Default '/dev/ttyUSB0'.")
    parser.add_argument('--unit', '-u', default='A', type=str, help="The "
                        "device unit ID, A-Z. Should only be used if multiple "
                        "flow controllers are connected to one port or if"
                        "device ID is not A.")
    parser.add_argument('--set-gas', '-g', default=None, type=str,
                        help="Sets the gas type. Supported gas types are: "
                             "'Air', 'Ar', 'CH4', 'CO', 'CO2', 'C2H6', 'H2', "
                             "'He', 'N2', 'N2O', 'Ne', 'O2', 'C3H8', "
                             "'n-C4H10', 'C2H2', 'C2H4', 'i-C2H10', 'Kr', "
                             "'Xe', 'SF6', 'C-25', 'C-10', 'C-8', 'C-2', "
                             "'C-75', 'A-75', 'A-25', 'A1025', 'Star29', "
                             "'P-5'")
    parser.add_argument('--set-flow-rate', '-f', default=None, type=float,
                        help="Sets the target flow rate of the controller.")
    parser.add_argument('--set-pressure', '-p', default=None, type=float,
                        help="Sets the target pressure of the controller.")
    parser.add_argument('--stream', '-s', action='store_true',
                        help="Sends a constant stream of flow controller "
                             "data, formatted as a tab-separated table.")
    parser.add_argument("--lock", "-l", action="store_true",
                        help="Locks device display.")
    parser.add_argument("--unlock", "-ul", action="store_true",
                        help="Unlocks device display.")
    parser.add_argument("--hold", "-hd", action="store_true",
                        help="Holds the valve at the present value.")
    parser.add_argument("--cancel-hold", "-c", action="store_true",
                        help="Cancel valve hold.")
    parser.add_argument("--reset-totalizer", "-r", action="store_true",
                        help="Reset current value of totalizer to zero.")
    args = parser.parse_args(args)
    async def get() -> None:
        async with FlowController(address=args.address, unit=args.unit) as flow_controller:
            if args.set_gas:
                await flow_controller.set_gas(args.set_gas)
            if args.set_flow_rate is not None and args.set_pressure is not None:
                raise ValueError("Cannot set both flow rate and pressure.")
            if args.set_flow_rate is not None:
                await flow_controller.set_flow_rate(args.set_flow_rate)
            if args.set_pressure is not None:
                await flow_controller.set_pressure(args.set_pressure)
            if args.lock:
                await flow_controller.lock()
            if args.unlock:
                await flow_controller.unlock()
            if args.hold:
                await flow_controller.hold()
            if args.cancel_hold:
                await flow_controller.cancel_hold()
            if args.reset_totalizer:
                await flow_controller.reset_totalizer()
            state = await flow_controller.get()
            if args.stream:
                try:
                    print('time\t' + '\t'.join(flow_controller.keys))
                    t0 = time()
                    while True:
                        state = await flow_controller.get()
                        print(f'{time() - t0:.2f}\t' +
                              '\t\t'.join(f'{state[key]:.2f}'
                                          for key in flow_controller.keys[:-1]) +
                              '\t\t' + state['gas'])
                except KeyboardInterrupt:
                    pass
            else:
                print(json.dumps(state, indent=2, sort_keys=True))
            await flow_controller.close()

    asyncio.run(get())


if __name__ == '__main__':
    command_line()
